from yankee.base.deserializer import DefaultAccessor
from yankee.util import do_nothing, camelize

class JsonMixin(object):
    def make_accessor(self):
        if self.data_key == False:
            return do_nothing
        data_key = self.data_key or self.name
        if data_key is None:
            return do_nothing
        return DefaultAccessor(camelize(data_key), many=self.many, filter=self.filter)



from yankee.util import do_nothing
from collections.abc import Mapping, Sequence


class JsonPath():
    def __init__(self, data_key=None):
        self.key = data_key.split(".")

    def __call__(self, obj):
        try:
            result = obj
            for k in self.key:
                if isinstance(result, Sequence) and k.isdigit():
                    result = result[int(k)]
                elif isinstance(result, Mapping):
                    result = result[k]
                else:
                    result = getattr(result, k)
            return result
        except (AttributeError, KeyError, IndexError):
            return None


class DefaultAccessor():
    def __init__(self, data_key=None, many=False, filter=None):
        self.data_key = data_key
        self.many = many
        self.filter = filter
        self.path_obj = self.make_path_obj()

    def make_path_obj(self):
        return JsonPath(self.data_key) if self.data_key is not None else do_nothing

    def __call__(self, *args, **kwargs):
        result = self.path_obj(*args, **kwargs)
        result = self.apply_filter(result)
        if not self.many and isinstance(result, list):
            if len(result) == 1:
                result = result[0]
            elif len(result) == 0:
                result = None
        return self.raise_exceptions(result)
        
    def apply_filter(self, result):
        if self.filter is None:
            return result
        if isinstance(result, list):
            return [r for r in result if self.filter(r)]
        else:
            return result if self.filter(result) else None

    def raise_exceptions(self, result):
        if self.many and not isinstance(result, list):
            raise ValueError("Expected many results, got one!")
        elif not self.many and isinstance(result, list):
            raise ValueError("Expected one result, got many!")
        return result

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['path_obj']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.path_obj = self.make_path_obj()


class Deserializer(object):
    class Meta:
        pass

    def __init__(self, data_key=None, many=False, required=False, filter=None):
        self.data_key = data_key
        self.required = required
        self.filter = filter
        self.many = many
        self.bind()

    def bind(self, name=None, parent=None):
        self.name = name
        self.parent = parent
        if self.parent is not None:
            self.Meta = parent.Meta
        self.accessor = self.make_accessor()
        return self

    def make_accessor(self):
        if self.data_key == False:
            return do_nothing
        key = self.data_key or self.name
        return DefaultAccessor(key, many=self.many, filter=self.filter)

    def load(self, obj):
        pre_obj = self.pre_load(obj)
        plucked_obj = self.get_obj(pre_obj)
        loaded_obj = self.deserialize(plucked_obj)
        return self.post_load(loaded_obj)

    def pre_load(self, obj):
        return obj
    
    def deserialize(self, obj):
        if obj is None and self.required:
            raise ValueError(
                f"Field {self.name} is required! Key {self.key} not found in {obj}"
            )
        return obj

    def post_load(self, obj):
        return obj

    def get_obj(self, obj):
        if self.many:
            if obj is None:
                return list()
            plucked_obj = self.accessor(obj)
            if not isinstance(plucked_obj, list):
                raise ValueError("Key Function returned single result rather than list!")
            return plucked_obj
        else:
            if obj is None:
                return None
            plucked_obj = self.accessor(obj)
            if isinstance(plucked_obj, list):
                raise ValueError("Key Function returned list of results!")
            return plucked_obj



from ...io.iterparse import file_iterparse


def xml_iterparse(file_obj: "io.RawBytesIO", tag=None):
    if not tag:
        return
    if isinstance(tag, bytes):
        tag = str(tag)
    start = f"<{tag}".encode("utf-8")
    end = f"(</{tag}>|{tag}.*?/>)".encode("utf-8")
    yield from file_iterparse(file_obj, start=start, end=end)



import datetime

import pytest

from yankee.json import Schema
from yankee.json import fields as f

doc = {
    "string": "Some String Data",
    "dateTime": "2021-05-04T12:05",
    "date": "2021-05-04",
    "booleans": ["True", "true", "False", "false"],
    "float": 1.234,
    "int": 23,
    "exists": "Something",
    "name": {"part1": "George", "part2": "Burdell"},
    "random": "Some data",
    "addressLine1": "1234 Anywhere",
    "addressLine2": "Austin, TX 71234",
}

class AddressField(f.Combine):
    addressLine1 = f.Str()
    addressLine2 = f.Str()

    def combine_func(self, obj):
        return f"{obj['addressLine1']}\n{obj['addressLine2']}"

class NameSchema(f.Combine):
    part1 = f.Str()
    part2 = f.Str()

    def combine_func(self, obj):
        return f"{obj['part1']} {obj['part2']}"

class SubSchema(Schema):
    string = f.Str()
    float = f.Float()

class ExampleSchema(Schema):
    string = f.Str()
    date_time = f.DT()
    date = f.Date()
    booleans = f.List(f.Bool, data_key="booleans")
    float = f.Float()
    int = f.Int()
    exists = f.Exists()
    does_not_exist = f.Exists()
    name = NameSchema()
    sub = SubSchema(False)
    address = AddressField(False)
    


def test_fields():
    schema = ExampleSchema()
    data = schema.load(doc)
    assert data["string"] == "Some String Data"
    assert data["date_time"] == datetime.datetime(2021, 5, 4, 12, 5)
    assert data["date"] == datetime.date(2021, 5, 4)
    assert data["booleans"] == [True, True, False, False]
    assert data["float"] - 1.234 < 0.001
    assert data["int"] == 23
    assert data["exists"] == True
    assert data["does_not_exist"] == False
    assert data["name"] == "George Burdell"
    assert data['sub']['string'] == "Some String Data"
    assert data['address'] == "1234 Anywhere\nAustin, TX 71234"

def test_filter_field():
    doc = {"list": [
        "el1",
        "el2"
    ]}

    class ExampleSchema(Schema):
        el_type = f.Str("list", filter=lambda o: o == "el2")
    
    result = ExampleSchema().load(doc)
    assert result['el_type'] == "el2"

    


from __future__ import annotations

import datetime
import re
import copy

from dateutil.parser import parse as parse_dt, isoparse

from yankee.util import AttrDict, clean_whitespace, is_valid

from .deserializer import Deserializer
from .schema import Schema


class Field(Deserializer):
    pass

class String(Field):
    def __init__(self, *args, formatter=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.formatter = formatter or clean_whitespace

    def deserialize(self, elem) -> "Optional[str]":
        elem = super().deserialize(elem)
        if elem is None or elem == "":
            return None
        else:
            return self.formatter(self.to_string(elem))

    def to_string(self, elem): # Abstracted out since XML requires a function call
        return str(elem)


class DateTime(String):
    def __init__(self, *args, dt_format=None, **kwargs):
        super().__init__(*args, **kwargs)
        if dt_format:
            self.parse_date = lambda s: datetime.datetime.strptime(s, dt_format)

    def parse_date(self, text:str):
        try:
            return isoparse(text)
        except ValueError:
            return parse_dt(text)

    def deserialize(self, elem) -> "Optional[datetime.datetime]":
        string = super(DateTime, self).deserialize(elem)
        if not string:
            return None
        return self.parse_date(string)      



class Date(DateTime):
    def deserialize(self, elem) -> "Optional[datetime.date]":
        date_time = super().deserialize(elem)
        return date_time.date() if date_time else None


class Boolean(String):
    def __init__(
        self, *args, true_value="true", case_sensitive=False, allow_none=True, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.true_value = true_value
        self.case_sensitive = case_sensitive
        self.allow_none = allow_none
        if not self.case_sensitive:
            self.true_value = self.true_value.lower()

    def deserialize(self, elem) -> "Optional[bool]":
        string = super(Boolean, self).deserialize(elem)
        if string is None or string == '':
            return None if self.allow_none else False
        if not self.case_sensitive:
            string = string.lower()
        return string == self.true_value


class Float(String):
    def deserialize(self, elem) -> "Optional[float]":
        string = super(Float, self).deserialize(elem)
        return float(string) if string is not None else None


class Integer(String):
    def deserialize(self, elem) -> "Optional[int]":
        string = super(Integer, self).deserialize(elem)
        return int(string) if string is not None else None


class Exists(Field):
    def deserialize(self, elem) -> bool:
        obj = super(Exists, self).deserialize(elem)
        return obj is not None


class Const(Field):
    def __init__(self, const, *args, **kwargs):
        self.const = const

    def deserialize(self, elem) -> "Any":
        return self.const


# Multiple Value Fields
class List(Field):
    def __init__(self, item_schema, data_key=None, **kwargs):
        kwargs['many'] = True
        self.item_schema = item_schema
        if callable(self.item_schema):
            self.item_schema = item_schema()
        super().__init__(data_key, **kwargs)

    def bind(self, name=None, schema=None):
        super().bind(name, schema)
        self.item_schema.bind(None, schema)

    def deserialize(self, obj):
        obj_gen = (self.item_schema.load(i) for i in obj)
        return [o for o in obj_gen if is_valid(o)]

class Dictionary(List):
    """Converts a list of items into a dictionary based on
    the key and value fields passed to it.
    """
    
    def __init__(self, data_key, key:Field, value:Field, **kwargs):
        self.key = key
        self.value = value
        return super().__init__(Field(), data_key, **kwargs)
    
    def deserialize(self, obj):
        obj = super().deserialize(obj)
        return AttrDict((self.key.load(i), self.value.load(i)) for i in obj)

# String Parsing Fields

class DelimitedString(String):
    def __init__(self, item_schema, data_key=None, delimeter=",", **kwargs):
        self.item_schema = item_schema
        if not isinstance(delimeter, re.Pattern):
            delimeter = re.compile(delimeter)
        self.delimeter = delimeter
        if callable(self.item_schema):
            self.item_schema = item_schema()
        super().__init__(data_key=data_key, **kwargs)

    def bind(self, name=None, schema=None):
        super().bind(name, schema)
        self.item_schema.bind(None, schema)

    def deserialize(self, obj):
        obj = super().deserialize(obj)
        if obj is None:
            return None
        objs = (self.item_schema.load(o) for o in self.delimeter.split(obj))
        return [o for o in objs if is_valid(o)]



# Schema-Like Fields

class Combine(Schema):
    """Can have fields like a schema that are then
    passed as an object to a combine function that
    transforms it to a single string value"""
 

    def get_output_name(self, name):
        return name

    def combine_func(self, obj):
        raise NotImplementedError("Must be implemented in subclass")

    def deserialize(self, raw_obj) -> "Optional[str]":
        obj = super().deserialize(raw_obj)
        return self.combine_func(obj)


class Alternative(Schema):
    """There may be a piece of data that has different names
    in different contexts. This has fields like a schema, then
    passes as a value the first non-empty or non-null result"""

    def deserialize(self, et_elem):
        obj = super().deserialize(et_elem)
        return next((v for v in obj.values() if is_valid(v)), None)


class ZipSchema(Schema):
    _list_field = List
    """Sometimes data is provided as a bunch of arrays, like:
    {
        "name": ["Peter", "Parker"],
        "age": [15, 25],
    }
    and we want to build out complete records from this data.
    This field performs that step:
    """

    def bind(self, name=None, parent=None):
        super().bind(name, parent)
        zip_fields = dict()
        for k, v in self.fields.items():
            v_copy = copy.deepcopy(v)
            # Remove the data key
            data_key = v_copy.data_key
            v_copy.data_key = None
            # Place it on the list field
            zip_fields[k] = self._list_field(v_copy, data_key)
        self.unzip_fields = self.fields
        self.fields = zip_fields

    def deserialize(self, obj) -> "Dict":
        objs = super().deserialize(obj)
        return self.lists_to_records(objs)
    
    def lists_to_records(self, obj):
        keys = tuple(obj.keys())
        values = tuple(obj.values())
        return [dict(zip(keys, v)) for v in zip(*values)]

# Aliases
Str = String
DT = DateTime
Bool = Boolean
Int = Integer
Alt = Alternative
Dict = Dictionary


from yankee.base import fields

from .key import XmlMixin


class Field(XmlMixin, fields.Field):
    pass


class String(XmlMixin, fields.String):
    pass


class DateTime(XmlMixin, fields.DateTime):
    pass


class Date(XmlMixin, fields.Date):
    pass


class Boolean(XmlMixin, fields.Boolean):
    pass


class Float(XmlMixin, fields.Float):
    pass


class Integer(XmlMixin, fields.Integer):
    pass


class Exists(XmlMixin, fields.Exists):
    pass


class Const(XmlMixin, fields.Const):
    pass


class List(XmlMixin, fields.List):
    pass

class Dict(XmlMixin, fields.Dict):
    pass

class Combine(XmlMixin, fields.Combine):
    pass


class Alternative(XmlMixin, fields.Alternative):
    pass


class ZipSchema(XmlMixin, fields.ZipSchema):
    _list_field = List

class DelimitedString(XmlMixin, fields.DelimitedString):
    pass


# Aliases
Str = String
DT = DateTime
Bool = Boolean
Int = Integer
Alt = Alternative



import lxml.etree as ET
from yankee.base.deserializer import DefaultAccessor

def do_nothing(obj):
    return obj

class XPathAccessor(DefaultAccessor):
    def __init__(self, *args, namespaces=dict(), **kwargs):
        self.namespaces = namespaces
        super().__init__(*args, **kwargs)

    def make_path_obj(self):
        if not self.many and not self.filter:
            xpath = f"({self.data_key})[1]"
        else:
            xpath = self.data_key
        return ET.XPath(xpath, namespaces=self.namespaces)


class XmlMixin(object):
    def make_accessor(self):
        namespaces = getattr(self.Meta, "namespaces", dict())
        if self.data_key is None:
            return do_nothing
        return XPathAccessor(self.data_key, many=self.many, filter=self.filter, namespaces=namespaces)

    def to_string(self, elem):
        if isinstance(elem, str):
            return elem
        try:
            return "".join(elem.itertext())
        except ValueError:
            # Occasionally elements won't have itertext available
            return elem.text

    def convert_groupdict(self, dictionary):
        root = ET.Element("root")
        for k, v in dictionary.items():
            subel = ET.SubElement(root, k)
            subel.text = v
        return root
